------------------------------------------------------------------------
-- SuccessionDKP
------------------------------------------------------------------------
-- An extension of WebDKP allowing you to use percentage-based DKP
-- model, while retaining the ability to sync with WebDKP.com
--
--
-- HOW THIS ADDON IS ORGANIZED:
-- The addon is grouped into a series of files which hold code for
-- specific functions. 
--
-- Core.lua       Core functions related to managing the addon
--
-- Whispers.lua   Handling various whisper functions (incl. bidding)
--
-- WebDKP.lua     Interaction with WebDKP
--
-- TODO:
--   * Fix bug where DKP changes in the middle of looting
--   * Incorporate rank checks
--   * Add offspec/ilvl bids
--   * Handle 2 of the same item dropping
--   * Ignore Emblems, Orbs, Fragments
------------------------------------------------------------------------

SuccessionDKP = LibStub("AceAddon-3.0"):NewAddon("SuccessionDKP", "AceConsole-3.0", "AceEvent-3.0")

local options = {
   name = "SuccessionDKP",
   handler = SuccessionDKP,
   type = "group",
   args = {
      lootThreshold = {
         type = "select",
         name = "Loot Threshold",
         desc = "Minimum threshold for item loots",
         get = function(info) return tostring(SuccessionDKP.db.profile.lootThreshold) end,
         set = function(info, val) SuccessionDKP.db.profile.lootThreshold = val end,
         values = {
            ["0"] = "|cff9d9d9dPoor|r",
            ["1"] = "|cffffffffCommon|r",
            ["2"] = "|cff1eff00Uncommon|r",
            ["3"] = "|cff0070ddRare|r",
            ["4"] = "|cffa335eeEpic|r",
            ["5"] = "|cffff8000Legendary|r",
         },
         order = 0,
      },
      lootAnnounceChan = {
         type = "select",
         name = "Announce Channel",
         desc = "The channel in which to announce items looted",
         get = function(info) return tostring(SuccessionDKP.db.profile.lootAnnounceChan) end,
         set = function(info, val) SuccessionDKP.db.profile.lootAnnounceChan = val end,
         values = {
            ["SAY"] = "Say",
            ["EMOTE"] = "Emote",
            ["YELL"] = "Yell",
            ["PARTY"] = "Party",
            ["GUILD"] = "Guild",
            ["OFFICER"] = "Officer",
            ["RAID"] = "Raid",
            ["RAID_WARNING"] = "Raid Warning"
         },
         order = 1,
      },
      announceLoot = {
         type = "execute",
         name = "Announce Loot",
         desc = "Announces the loot of the boss",
         func = "announceLoot",
         order = 2,
      },
      printBidders = {
         type = "execute",
         name = "Print Bidders",
         desc = "Print bidders for each item",
         func = "printBidders",
         order = 3,
      },
      done = {
         type = "execute",
         name = "Done Looting",
         desc = "Click this button when all the loot has been given out.",
         func = "doneBidding",
         order = 4,
         width = "double",
      },
      winners = {
         type = "execute",
         name = "Announce Winners",
         desc = "Announces the winners for all the items",
         func = "announceWinners",
         order = 5,
         width = "double",
      },
   },
}

local defaultOptions = {
   profile = {
      lootThreshold = 3,
      lootAnnounceChan = "RAID",
      
      isBidding = false,
      
      lootedItems = {}, -- An array of the items being distributed
      bids = {},
      didBid = {},
   },
}
-- Called when the addon is loaded
function SuccessionDKP:OnInitialize()
   -- Setup our local database
   self.db = LibStub("AceDB-3.0"):New("SuccessionDKPDB", defaultOptions, "Default")
   
   -- Register our option table (see 'options' above)
   LibStub("AceConfig-3.0"):RegisterOptionsTable("SuccessionDKP", options)
   
   -- Add our options to the Blizzard's addon options menu
   self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SuccessionDKP", "SuccessionDKP")
   
   -- Register the chat commands to a ChatCommand function
   self:RegisterChatCommand("successiondkp", "ChatCommand")
   self:RegisterChatCommand("sdkp", "ChatCommand")
end

-- Called when the addon is enabled
function SuccessionDKP:OnEnable()
   -- Register to listen for the loot event
   SuccessionDKP:RegisterEvent("LOOT_OPENED")
   
   -- Listen for whispers
   SuccessionDKP:RegisterEvent("CHAT_MSG_WHISPER")
   
end

-- Called when the addon is disabled
function SuccessionDKP:OnDisable()
   
end

-- Called when the user inputs a slash command
function SuccessionDKP:ChatCommand(input)
   if (not input) or (input:trim() == "") then
      -- If we don't have input, open the addon GUI
      InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
   else
      -- Otherwise, handle the slash command
      LibStub("AceConfigCmd-3.0").HandleCommand(SuccessionDKP, "sdkp", "SuccessionDKP", input)
   end
end

function SuccessionDKP:LOOT_OPENED()

   -- Don't care about new loot if we're in the middle of bidding
   if (isBidding == true) then
      return
   end
   
   self:Print("Bidding session started.");

   -- Reset the tables
   table.wipe(self.db.profile.lootedItems)
   table.wipe(self.db.profile.bids)
   table.wipe(self.db.profile.didBid)
   
   local lootIndex = 1
   
   for index = 1, GetNumLootItems() do
      if (LootSlotIsItem(index)) then
      -- The loot is an item
         local _, _, quantity = GetLootSlotInfo(index)
         local _, _, rarity = GetItemInfo(GetLootSlotLink(index))

         if (rarity >= tonumber(self.db.profile.lootThreshold)) then
            if (quantity == 1) then
               self.db.profile.lootedItems[lootIndex] = GetLootSlotLink(index)
            else
               self.db.profile.lootedItems[lootIndex] = GetLootSlotLink(index) .. "x" .. quantity
            end
            self.db.profile.bids[lootIndex] = {}
            self.db.profile.didBid[lootIndex] = {}
            lootIndex = lootIndex + 1
         end
         
      else
         -- The loot is coins
      end
   end
   
   isBidding = true
end

function SuccessionDKP:announceLoot()
   for index = 1, table.getn(self.db.profile.lootedItems) do
      SendChatMessage(index .. ")  " .. self.db.profile.lootedItems[index], self.db.profile.lootAnnounceChan)
   end
end

function SuccessionDKP:announceWinners()
   local dkpCost, iLvl, iName, tieIndex, maxId
   local tieRollTable = {}
   for lootIndex = 1, table.getn(self.db.profile.lootedItems) do
      iName, _, _, iLvl = GetItemInfo(self.db.profile.lootedItems[lootIndex])
      if (iLvl < 226) then iLvl = 226 end
      if (iName:sub(-9) == "Conqueror") then iLvl = 229 end
      if (iName:sub(-10) == "Vanquisher") then iLvl = 229 end
      if (iName:sub(-9) == "Protector") then iLvl = 229 end
      
      if (table.getn(self.db.profile.bids[lootIndex]) == 0) then
         SendChatMessage(self.db.profile.lootedItems[lootIndex] .. " rots.", self.db.profile.lootAnnounceChan)
      else
         dkpCost = tonumber(self.db.profile.bids[lootIndex][1]:sub(1,7)) * ((iLvl - 226) * 0.03 + 0.18)
         if (table.getn(self.db.profile.bids[lootIndex]) > 1) then
            if (self.db.profile.bids[lootIndex][1]:sub(1,7) == self.db.profile.bids[lootIndex][2]:sub(1,7)) then
               -- We have a tie
               SendChatMessage("TIE for " .. self.db.profile.lootedItems[lootIndex] .. ". Rolling...", self.db.profile.lootAnnounceChan)
               table.wipe(tieRollTable)
               for tieIndex = 1, table.getn(self.db.profile.bids[lootIndex]) do
                  if (self.db.profile.bids[lootIndex][1]:sub(1,7) == self.db.profile.bids[lootIndex][tieIndex]:sub(1,7)) then
                     table.insert(tieRollTable, string.format("%02.0f", math.random(1,100)))
                     SendChatMessage("  " .. self.db.profile.bids[lootIndex][tieIndex]:sub(11) .. " rolls a " .. string.format("%g", tieRollTable[tieIndex]) .. " (1-100)", self.db.profile.lootAnnounceChan)
                  end
               end
               
               maxId = 1
               for tieIndex = 2, table.getn(tieRollTable) do
                  if (tieRollTable[tieIndex] > tieRollTable[maxId]) then
                     maxId = tieIndex
                  end
               end
               
               SendChatMessage(self.db.profile.bids[lootIndex][maxId]:sub(11) .. " wins " .. self.db.profile.lootedItems[lootIndex] .. " for " .. dkpCost .. "dkp!", self.db.profile.lootAnnounceChan)
            else
               -- We have a winner (highest DKP)
               SendChatMessage(self.db.profile.bids[lootIndex][1]:sub(11) .. " wins " .. self.db.profile.lootedItems[lootIndex] .. " for " .. dkpCost .. "dkp!", self.db.profile.lootAnnounceChan)
            end
         else
            -- We have a winner (no other bidders)
            SendChatMessage(self.db.profile.bids[lootIndex][1]:sub(11) .. " wins " .. self.db.profile.lootedItems[lootIndex] .. " for " .. dkpCost .. "dkp!", self.db.profile.lootAnnounceChan)
         end
      end
   end
end

function SuccessionDKP:printBidders()
      local dkpCost
      
      for itemIndex = 1, table.getn(self.db.profile.lootedItems) do
         self:Print(itemIndex .. ")  " .. self.db.profile.lootedItems[itemIndex])
         for bidderIndex = 1, table.getn(self.db.profile.bids[itemIndex]) do
             self:Print("  " .. self.db.profile.bids[itemIndex][bidderIndex])
         end
      end
   end

function SuccessionDKP:CHAT_MSG_WHISPER(event, msg, author, _, status, _, _)
   local spaceSpot = string.find(msg, " ")
   local itemNumber = tonumber(string.sub(msg, 1, 1))
   
   -- Check to see if we have a valid item number
   if (spaceSpot == 2) then 
      if ((itemNumber >= 1) and (itemNumber <= table.getn(self.db.profile.lootedItems))) then      
      
         if ( WebDKP_DkpTable[author] == nil ) then
            SendChatMessage("You cannot bid on an item unless you are in the raid.", "WHISPER", nil, author)
            return
         end
      
         -- Get the bid string
         local bidString = string.format("%07.2f", WebDKP_DkpTable[author]["dkp_1"]) .. " - " .. author
                     
         -- Check to see if the bid amount is valid
         if (string.upper(string.sub(msg, 3, 5)) == "BID") then
            if (self.db.profile.didBid[itemNumber][author] == true) then
               SendChatMessage("You have already bid on " .. self.db.profile.lootedItems[itemNumber] .. ".", "WHISPER", nil, author)
             else
               -- max bid
               self:Print(author .. " bid on " .. self.db.profile.lootedItems[itemNumber] .. ".")
               SendChatMessage("You have bid on " .. self.db.profile.lootedItems[itemNumber], "WHISPER", nil, author)
               SendChatMessage("To retract your bid, reply with '" .. itemNumber .. " retract'.", "WHISPER", nil, author)
               table.insert(self.db.profile.bids[itemNumber], bidString)
               self.db.profile.didBid[itemNumber][author] = true
               table.sort(self.db.profile.bids[itemNumber], function (a, b) return a > b end)
             end
             return
         end

         if (string.upper(string.sub(msg, 3, 5)) == "RET") then
            if (self.db.profile.didBid[itemNumber][author]) then
               for key, value in pairs(self.db.profile.bids[itemNumber]) do
                  if (value:sub(11) == author) then
                     SendChatMessage("Retracted your bid for " .. self.db.profile.lootedItems[itemNumber] .. ".", "WHISPER", nil, author)
                     table.remove(self.db.profile.bids[itemNumber], key)
                  end
               end
               self.db.profile.didBid[itemNumber][author] = false
            else
               SendChatMessage("You are not currently bidding on " .. self.db.profile.lootedItems[itemNumber] .. ".", "WHISPER", nil, author)
            end
            return 
         end 
         
         -- Invalid bid amount
         SendChatMessage("Invalid bid amount: " .. string.sub(msg, 3, 5) .. ". Valid amounts: BID, RETRACT.", "WHISPER", nil, author)
      else
         -- Invalid item number
         SendChatMessage("Invalid item: " .. itemNumber, "WHISPER", nil, author);
      end
   end
end

function round(num, idp)
  local mult = 10^(idp or 0)
  return math.floor(num * mult + 0.5) / mult
end

function SuccessionDKP:doneBidding()
   self:Print("Bidding session ended.")
   isBidding = false
end